from __future__ import annotations

import re
import shlex
from typing import List

from huggingface_hub.utils import validate_repo_id

from paka.cluster.context import Context
from paka.cluster.utils import get_model_store
from paka.config import CloudModelGroup
from paka.constants import MODEL_MOUNT_PATH
from paka.k8s.utils import get_gpu_count


# Heuristic to determine if the image is a vLLM image
def is_vllm_image(image: str) -> bool:
    return image.lower().startswith("vllm")


def get_runtime_command_vllm(ctx: Context, model_group: CloudModelGroup) -> List[str]:
    runtime = model_group.runtime
    if runtime.command:
        command_str = " ".join(runtime.command) if runtime.command else ""
        if re.search(r"(--model)[ \t]*\S+", command_str):
            return runtime.command

    if model_group.model:
        if model_group.model.useModelStore:
            store = get_model_store(ctx, with_progress_bar=False)
            if not store.glob(f"{model_group.name}/*"):
                raise ValueError(
                    f"No model named {model_group.name} was found in the model store."
                )
            model_to_load = f"{MODEL_MOUNT_PATH}"
        elif model_group.model.hfRepoId:
            validate_repo_id(model_group.model.hfRepoId)
            model_to_load = model_group.model.hfRepoId
        else:
            raise ValueError("Did not find a model to load.")

    def attach_model_to_command(command: List[str]) -> List[str]:
        return command + ["--model", model_to_load]

    if runtime.command:
        return attach_model_to_command(runtime.command)

    command = shlex.split(
        f"python3 -O -u -m vllm.entrypoints.openai.api_server --host 0.0.0.0 --served-model-name {model_group.name}"
    )

    gpu_count = get_gpu_count(ctx, model_group)

    if gpu_count > 1:
        command += ["--tensor-parallel-size", str(gpu_count)]

    return attach_model_to_command(command)



import struct
from collections import namedtuple
from typing import Any, BinaryIO, Dict, List, Union

Slice = namedtuple("Slice", ["value", "length"])

GGUFValueType = {
    "UINT8": 0,
    "INT8": 1,
    "UINT16": 2,
    "INT16": 3,
    "UINT32": 4,
    "INT32": 5,
    "FLOAT32": 6,
    "BOOL": 7,
    "STRING": 8,
    "ARRAY": 9,
    "UINT64": 10,
    "INT64": 11,
    "FLOAT64": 12,
}


def read_string(file: BinaryIO, version: int, little_endian: bool) -> Slice:
    length = read_versioned_size(file, version, little_endian)
    value = file.read(length.value).decode("utf-8")
    return Slice(value=value, length=length.length + length.value)


def read_versioned_size(file: BinaryIO, version: int, little_endian: bool) -> Slice:
    endian = "<" if little_endian else ">"
    if version == 1:
        length_value = struct.unpack(endian + "I", file.read(4))[0]
        length_size = 4
    elif version in [2, 3]:
        length_value = struct.unpack(endian + "Q", file.read(8))[0]
        length_size = 8
    else:
        raise ValueError(f"Unsupported version: {version}")
    return Slice(value=length_value, length=length_size)


def read_metadata_value(
    file: BinaryIO, value_type: int, version: int, little_endian: bool
) -> Slice:
    endian = "<" if little_endian else ">"
    if value_type == GGUFValueType["UINT8"]:
        return Slice(value=struct.unpack(endian + "B", file.read(1))[0], length=1)
    elif value_type == GGUFValueType["INT8"]:
        return Slice(value=struct.unpack(endian + "b", file.read(1))[0], length=1)
    elif value_type == GGUFValueType["UINT16"]:
        return Slice(value=struct.unpack(endian + "H", file.read(2))[0], length=2)
    elif value_type == GGUFValueType["INT16"]:
        return Slice(value=struct.unpack(endian + "h", file.read(2))[0], length=2)
    elif value_type == GGUFValueType["UINT32"]:
        return Slice(value=struct.unpack(endian + "I", file.read(4))[0], length=4)
    elif value_type == GGUFValueType["INT32"]:
        return Slice(value=struct.unpack(endian + "i", file.read(4))[0], length=4)
    elif value_type == GGUFValueType["FLOAT32"]:
        return Slice(value=struct.unpack(endian + "f", file.read(4))[0], length=4)
    elif value_type == GGUFValueType["BOOL"]:
        return Slice(value=struct.unpack(endian + "B", file.read(1))[0] != 0, length=1)
    elif value_type == GGUFValueType["STRING"]:
        return read_string(file, version, little_endian)
    elif value_type == GGUFValueType["ARRAY"]:
        array_type = struct.unpack(endian + "I", file.read(4))[0]
        array_length = read_versioned_size(file, version, little_endian)
        length = 4 + array_length.length
        array_values = []
        for _ in range(array_length.value):
            metadata_value = read_metadata_value(
                file, array_type, version, little_endian
            )
            array_values.append(metadata_value.value)
            length += metadata_value.length
        return Slice(value=array_values, length=length)
    elif value_type == GGUFValueType["UINT64"]:
        return Slice(value=struct.unpack(endian + "Q", file.read(8))[0], length=8)
    elif value_type == GGUFValueType["INT64"]:
        return Slice(value=struct.unpack(endian + "q", file.read(8))[0], length=8)
    elif value_type == GGUFValueType["FLOAT64"]:
        return Slice(value=struct.unpack(endian + "d", file.read(8))[0], length=8)
    else:
        raise ValueError(f"Unsupported metadata type: {value_type}")


def gguf(local_file_path: str) -> Dict[str, Any]:
    with open(local_file_path, "rb") as file:
        # Read and check the magic number
        magic_number = file.read(4)
        if magic_number != b"GGUF":
            raise ValueError(
                "Not a valid gguf file: does not start with GGUF magic number"
            )

        # Determine the endianness and read the version
        version = struct.unpack("<I", file.read(4))[0]
        if version & 65535:
            little_endian = True
        else:
            little_endian = False
            file.seek(4, 0)  # Rewind to read the version again in big-endian
            version = struct.unpack(">I", file.read(4))[0]

        if version not in [1, 2, 3]:
            raise ValueError(f"Not a valid gguf file: unsupported version '{version}'")

        # Read the tensor count and key-value count
        tensor_count_slice = read_versioned_size(file, version, little_endian)
        tensor_count = tensor_count_slice.value

        kv_count_slice = read_versioned_size(file, version, little_endian)
        kv_count = kv_count_slice.value

        # Initialize the metadata
        metadata: Dict[str, Any] = {
            "version": version,
            "tensor_count": tensor_count,
            "kv_count": kv_count,
        }

        for _ in range(kv_count):
            key_result = read_string(file, version, little_endian)
            key = key_result.value

            value_type = struct.unpack("<I" if little_endian else ">I", file.read(4))[0]
            if value_type not in GGUFValueType.values():
                raise ValueError(f"Unsupported metadata type: {value_type}")

            value_result = read_metadata_value(file, value_type, version, little_endian)
            metadata[key] = value_result.value

        tensor_infos: List[Dict[str, Union[str, int, List[int]]]] = []
        for _ in range(tensor_count):
            name_result = read_string(file, version, little_endian)
            name = name_result.value

            n_dims = struct.unpack("<I" if little_endian else ">I", file.read(4))[0]
            shape: List[int] = []
            for _ in range(n_dims):
                shape_dim = read_versioned_size(file, version, little_endian)
                shape.append(shape_dim.value)

            dtype = struct.unpack("<I" if little_endian else ">I", file.read(4))[0]
            offset = struct.unpack("<Q" if little_endian else ">Q", file.read(8))[0]

            tensor_infos.append(
                {
                    "name": name,
                    "n_dims": n_dims,
                    "shape": shape,
                    "dtype": dtype,
                    "offset": offset,
                }
            )

        return {"metadata": metadata, "tensor_infos": tensor_infos}


result = gguf("/Users/jleng/Downloads/llama-2-7b-chat.Q4_0.gguf")
metadata, tensor_infos = result["metadata"], result["tensor_infos"]

for k, v in metadata.items():
    if not k.startswith("tokenizer"):
        print(k, v)

print(tensor_infos)



import pulumi_aws as aws

from paka.cluster.context import Context
from paka.utils import call_once


@call_once
def create_container_registry(ctx: Context) -> None:
    """
    Create a container registry in AWS ECR for storing Docker images.

    Returns:
        None
    """
    repository = aws.ecr.Repository(
        ctx.cluster_name,
        force_delete=True,
        image_tag_mutability="MUTABLE",
    )

    # Save the repository URL to the cluster data file
    repository.repository_url.apply(lambda url: ctx.set_registry(url))



from __future__ import annotations

import json
import os
import re
from datetime import datetime, timezone
from typing import List, Literal, Optional, Tuple

import click
import typer
from kubernetes.client.exceptions import ApiException
from tabulate import tabulate

from paka.cli.utils import (
    format_timedelta,
    get_cluster_namespace,
    load_kubeconfig,
    process_envs,
    resolve_image,
)
from paka.k8s.function.service import (
    create_knative_service,
    delete_knative_service,
    list_knative_revisions,
    list_knative_services,
    split_traffic_among_revisions,
)
from paka.logger import logger
from paka.utils import kubify_name

function_app = typer.Typer()

VALID_PERCENTAGE_RANGE = (0, 100)


def validate_traffic_split(split: str) -> Tuple[str, int]:
    """
    Validate a single traffic split string and return the revision and percentage.

    Args:
        split (str): The traffic split string in the format 'revision=percentage'.

    Returns:
        Tuple[str, int]: A tuple containing the revision and percentage.

    Raises:
        ValueError: If the input string is not in the expected format or the percentage is out of range.
    """
    if "=" not in split:
        raise ValueError(f"Invalid format, missing '=': {split}")

    revision, percent_str = split.split("=")
    if not percent_str.strip().isdigit():
        raise ValueError(f"Invalid format or non-numeric percentage: {split}")

    percent = int(percent_str.strip())
    if percent not in range(*VALID_PERCENTAGE_RANGE):
        raise ValueError(
            f"Traffic percentage out of valid range ({VALID_PERCENTAGE_RANGE[0]}-{VALID_PERCENTAGE_RANGE[1]}): {percent}"
        )

    return revision, percent


def process_traffic_splits(
    traffic_splits: List[str],
) -> Tuple[List[Tuple[str, int]], int]:
    total_traffic_percent = 0
    splits = []
    revisions = set()
    for split_str in traffic_splits:
        for split in re.split(r",\s*", split_str):
            revision, percent = validate_traffic_split(split)
            if revision in revisions:
                logger.error(f"Error: Duplicate revision '{revision}' provided.")
                raise typer.Exit(1)
            revisions.add(revision)
            splits.append((revision, percent))
            total_traffic_percent += percent

    return splits, total_traffic_percent


@function_app.command()
def deploy(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
    name: str = typer.Option(
        ...,
        "--name",
        help="The name of the function to be deployed.",
    ),
    entrypoint: str = typer.Option(
        "web",
        "--entrypoint",
        help="The entrypoint of the application. This refers to the command "
        "defined in the Procfile that will be executed.",
    ),
    source_dir: Optional[str] = typer.Option(
        None,
        "--source",
        help="The directory containing the source code of the application. If "
        "specified, a new Docker image will be built using the source code from "
        "this directory. A Dockerfile is not required because the build process "
        "uses Cloud Native's Buildpacks, which automatically detect and install "
        "dependencies.",
    ),
    image: Optional[str] = typer.Option(
        None,
        "--image",
        help="The name of the Docker image to deploy. If both an image and a "
        "source directory are provided, this image will be used and the source "
        "directory will be ignored.",
    ),
    min_instances: int = typer.Option(
        1,
        "--min-replicas",
        help="The minimum number of instances to be maintained at all times to "
        "handle incoming requests. The default value is 0.",
    ),
    max_instances: int = typer.Option(
        0,
        "--max-replicas",
        help="The maximum number of instances that can be created to handle peak "
        "load. The default value is set to 0, indicating that there is no upper "
        "limit on the number of instances that can be created.",
    ),
    scaling_metric: Literal["concurrency", "rps"] = typer.Option(
        "concurrency",
        "--scaling-metric",
        help="The performance metric that the scaling system should monitor to "
        "determine when to adjust the number of instances.",
        case_sensitive=True,
        click_type=click.Choice(["rps", "concurrency"]),
    ),
    metric_target: int = typer.Option(
        1,
        "--metric-target",
        help="The desired value for the chosen scaling metric. The system will "
        "adjust the number of active instances to attempt to reach this target.",
    ),
    scale_down_delay: str = typer.Option(
        "0s",
        "--scale-down-delay",
        help="The delay before the system scales down after a spike in traffic. "
        "This value must be a duration between 0 seconds and 1 hour, represented "
        "as a string (e.g., '0s', '1m', '1h').",
    ),
    env_vars: List[str] = typer.Option(
        [],
        "--env",
        help="Specify environment variables for the function in the format 'key=value'",
        show_default=False,
    ),
    resource_requests: Optional[str] = typer.Option(
        None,
        "--resource-requests",
        help='The resource requests for the function, in JSON format. For example: \'{"cpu": "100m", "memory": "128Mi"}\'',
    ),
    resource_limits: Optional[str] = typer.Option(
        None,
        "--resource-limits",
        help='The resource limits for the function, in JSON format. For example: \'{"cpu": "200m", "memory": "256Mi"}\'',
    ),
) -> None:
    """
    Deploy a function to a Knative service.
    """
    load_kubeconfig(cluster_name)
    resolved_image = resolve_image(cluster_name, image, source_dir)

    logger.info(f"Deploying {name}...")

    resource_requests_dict = (
        json.loads(resource_requests) if resource_requests else None
    )
    resource_limits_dict = json.loads(resource_limits) if resource_limits else None

    envs = process_envs(env_vars)

    create_knative_service(
        service_name=kubify_name(name),
        namespace=get_cluster_namespace(cluster_name),
        image=resolved_image,
        entrypoint=entrypoint,
        min_instances=min_instances,
        max_instances=max_instances,
        scaling_metric=(scaling_metric, str(metric_target)),
        scale_down_delay=scale_down_delay,
        envs=envs,
        resource_requests=resource_requests_dict,
        resource_limits=resource_limits_dict,
    )

    logger.info(f"Successfully deployed {name}")


@function_app.command()
def list(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
) -> None:
    """
    List all deployed functions.
    """
    load_kubeconfig(cluster_name)
    services = list_knative_services(get_cluster_namespace(cluster_name))

    if not services.items:
        logger.info("No functions found.")
        return

    table = [
        (
            svc.metadata.name,
            svc.status.url,
            svc.status.latestCreatedRevisionName,
            svc.status.latestReadyRevisionName,
            svc.status.conditions[0].status if len(svc.status.conditions) > 0 else "",
            svc.status.conditions[0].reason if len(svc.status.conditions) > 0 else "",
        )
        for svc in services.items
    ]
    logger.info(
        tabulate(
            table,
            headers=[
                "Name",
                "Endpoint",
                "LatestCreated",
                "LatestReady",
                "Ready",
                "Reason",
            ],
        )
    )


@function_app.command()
def list_revisions(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
    name: str = typer.Argument(
        ...,
        help="The name of the function to list revisions for. If not provided, list revisions for all services.",
    ),
) -> None:
    """
    List all deployed functions.
    """
    load_kubeconfig(cluster_name)
    revisions = list_knative_revisions(get_cluster_namespace(cluster_name), name)

    if not revisions:
        logger.info("No revisions found.")
        return

    table = []
    for revision in revisions:
        filtered_conditions = [
            condition
            for condition in revision.status.conditions
            if condition.get("reason") != "NoTraffic"
        ]

        sorted_conditions = sorted(
            filtered_conditions,
            key=lambda condition: condition.get("lastTransitionTime"),
            reverse=True,
        )
        latest_condition = sorted_conditions[0] if sorted_conditions else None

        true_conditions = sum(
            c.get("status") == "True" for c in revision.status.conditions
        )

        table.append(
            (
                revision.metadata.name,
                revision.metadata.labels["serving.knative.dev/service"],
                revision.traffic,
                revision.metadata.labels.get("tags", ""),
                revision.metadata.labels["serving.knative.dev/configurationGeneration"],
                format_timedelta(
                    datetime.now(timezone.utc)
                    - datetime.strptime(
                        revision.metadata.creationTimestamp, "%Y-%m-%dT%H:%M:%SZ"
                    ).replace(tzinfo=timezone.utc)
                ),
                f"{true_conditions} OK / {len(revision.status.conditions)}",
                latest_condition.get("status", ""),
                latest_condition.get("reason", ""),
                # latest_condition.get("message", ""),
            )
        )

    logger.info(
        tabulate(
            table,
            headers=[
                "Name",
                "Service",
                "Traffic",
                "Tags",
                "Generation",
                "Age",
                "Conditions",
                "Status",
                "Reason",
                # "Message",
            ],
        )
    )


@function_app.command()
def update_traffic(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
    name: str = typer.Argument(
        ...,
        help="The name of the function to update traffic for.",
    ),
    traffic_splits: List[str] = typer.Option(
        [],
        "--traffic",
        help="Specify traffic splits for each revision in the format 'revision=percentage'. "
        "Multiple splits can be provided.",
        show_default=False,
    ),
    latest_revision_traffic: int = typer.Option(
        0,
        "--latest-revision-traffic",
        help="The percentage of traffic to send to the latest revision.",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Automatic yes to prompts. Use this option to bypass the confirmation "
        "prompt and directly proceed with the operation.",
    ),
) -> None:
    """
    Update the traffic distribution among the revisions of a function.

    You can provide the traffic splits in two ways:

    1. Using multiple `--traffic` options, e.g.:
       `update_traffic --traffic revision1=50 --traffic revision2=30 --traffic revision3=20`

    2. Using a comma-separated list within a single `--traffic` option, e.g.:
       `update_traffic --traffic revision1=50,revision2=30,revision3=20`

    If the total traffic percentage is less than 100% and `--latest-revision-traffic` is not provided,
    the user will be prompted to confirm whether the remaining traffic should be assigned to the latest revision.
    """
    splits, total_traffic_percent = process_traffic_splits(traffic_splits)

    if total_traffic_percent + latest_revision_traffic > 100:
        logger.error("Total traffic percent should not exceed 100%")
        raise typer.Exit(1)

    if total_traffic_percent < 100 and latest_revision_traffic == 0:
        remaining_traffic = 100 - total_traffic_percent
        confirm = yes or typer.confirm(
            f"Assign remaining {remaining_traffic}% traffic to the latest revision?",
            default=True,
        )
        if confirm:
            latest_revision_traffic = remaining_traffic
        else:
            logger.error("Traffic distribution aborted by user.")
            raise typer.Abort()

    load_kubeconfig(cluster_name)
    logger.info(f"Updating traffic for function {name}")
    split_traffic_among_revisions(
        get_cluster_namespace(cluster_name),
        name,
        splits,
        latest_revision_traffic,
    )
    logger.info(f"Successfully updated traffic for service {name}")


@function_app.command()
def delete(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
    name: str = typer.Argument(
        ...,
        help="The name of the function to delete.",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Automatic yes to prompts. Use this option to bypass the confirmation "
        "prompt and directly proceed with the deletion.",
    ),
) -> None:
    """
    Delete a function.
    """
    if yes or typer.confirm(
        f"Are you sure you want to delete the function {name}?", default=False
    ):
        load_kubeconfig(cluster_name)
        logger.info(f"Deleting function {name}")
        try:
            delete_knative_service(
                kubify_name(name), get_cluster_namespace(cluster_name)
            )
            logger.info(f"Successfully deleted function {name}")
        except ApiException as e:
            if e.status == 404:
                logger.error(f"Function {name} not found.")
            else:
                raise



from __future__ import annotations

import re
from typing import Any, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from ruamel.yaml import YAML

from paka.utils import to_yaml

CONFIG_VERSION = "1.3"


class PakaBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


def validate_size(v: str, error_message: str = "Invalid size format") -> str:
    """
    Validates the format of the size field.

    Args:
        v (str): The value of the size field.
        error_message (str, optional): The error message to raise if the format of the input value is invalid. Defaults to "Invalid size format".

    Returns:
        str: The input value if validation is successful.

    Raises:
        ValueError: If the format of the input value is invalid. The error message is specified by the `error_message` parameter.
    """
    if not re.match(r"^\d+(Mi|Gi)$", v):
        raise ValueError(error_message)
    return v


class ResourceRequest(PakaBaseModel):
    """
    Represents the resource request for a container.
    """

    cpu: str = Field(..., description="The amount of CPU to request.")
    memory: str = Field(..., description="The amount of memory to request.")
    gpu: Optional[int] = Field(None, description="The number of GPUs to request.")

    @field_validator("cpu", mode="before")
    def validate_cpu(cls, v: str) -> str:
        """
        Validates the format of the cpu field.

        Args:
            v (str): The value of the cpu field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        if not re.match(r"^\d+(m)?$", v):
            raise ValueError("Invalid CPU format")
        return v

    @field_validator("memory", mode="before")
    def validate_memory(cls, v: str) -> str:
        """
        Validates the format of the memory field.

        Args:
            v (str): The value of the memory field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid memory format")

    @field_validator("gpu", mode="before")
    def validate_gpu(cls, v: Optional[int]) -> Optional[int]:
        """
        Validates the value of the gpu field.

        Args:
            v (Optional[int]): The value of the gpu field.

        Returns:
            Optional[int]: The input value if validation is successful.

        Raises:
            ValueError: If the value is less than 0.
        """
        if v is not None and v < 0:
            raise ValueError("GPU count cannot be less than 0")
        return v


class AwsGpuNodeConfig(PakaBaseModel):
    """
    Represents a configuration for an AWS GPU node.
    """

    enabled: bool = Field(False, description="Whether the GPU node is enabled.")

    # At least 40 GB is required for the disk size
    diskSize: Optional[int] = Field(
        40, description="The size of the disk for the GPU node in GB."
    )


class GcpGpuNodeConfig(PakaBaseModel):
    """
    Represents a Google Cloud Platform GPU node.
    """

    imageType: str = Field(..., description="The type of image used for the GPU node.")
    acceleratorType: str = Field(
        ..., description="The type of accelerator used for the GPU node."
    )
    acceleratorCount: int = Field(
        ..., description="The number of accelerators attached to the GPU node."
    )
    diskType: str = Field(..., description="The type of disk used for the GPU node.")
    diskSize: int = Field(
        ..., description="The size of the disk attached to the GPU node in GB."
    )


class CloudNode(PakaBaseModel):
    """
    Represents a node in the cloud cluster.
    """

    nodeType: str = Field(..., description="The type of the node.")
    diskSize: int = Field(
        20, description="The size of the disk attached to the node in GB."
    )


class AwsNode(CloudNode):
    """
    Represents an AWS cloud node configuration.
    """

    gpu: Optional[AwsGpuNodeConfig] = Field(
        None, description="The AWS GPU node configuration, if applicable."
    )


class Runtime(PakaBaseModel):
    """
    Represents a runtime for a model.
    """

    image: str = Field(..., description="The Docker image to use for the runtime.")
    command: Optional[List[str]] = Field(
        None, description="The command to run in the Docker container."
    )
    env: Optional[List[Dict[str, Any]]] = None
    readinessProbe: Optional[Dict[str, Any]] = None
    livenessProbe: Optional[Dict[str, Any]] = None
    volumeMounts: Optional[List[Dict[str, Any]]] = None


class Model(PakaBaseModel):
    """
    Represents a model.
    """

    hfRepoId: Optional[str] = Field(
        None, description="The HuggingFace repository ID for the model."
    )
    files: List[str] = Field(
        ["*"], description="The list of files to include from the repository."
    )
    useModelStore: bool = Field(
        True, description="Whether to save the model to a model store, such as s3."
    )


class Trigger(PakaBaseModel):
    """
    Represents a trigger.
    """

    type: str = Field(..., description="The type of the trigger.")
    metadata: Dict[str, str] = Field(
        ..., description="The metadata associated with the trigger."
    )


class CloudModelGroup(CloudNode):
    """
    Represents a group of cloud models.
    """

    name: str = Field(..., description="The name of the model group.")
    model: Optional[Model] = Field(
        None,
        description="The model to deploy in the model group. If None, runtime image is responsible for loading the model.",
    )
    runtime: Runtime = Field(..., description="The runtime for the model group.")

    resourceRequest: Optional[ResourceRequest] = Field(
        None,
        description="The resource request for the model group, specifying the amount of CPU and memory to request.",
    )

    autoScaleTriggers: Optional[List[Trigger]] = Field(
        None,
        description="""The list of triggers for auto-scaling.
    Triggers for autoscaling the model group. Trigger are not strongly typed, so we use a list of dictionaries to represent them.
    For example:

    autoScaleTriggers=[
        # CPU trigger example
        Trigger(
            type="cpu",
            metadata={
                "type": "Utilization",
                "value": "70"
            }
        ),
        # Prometheus trigger example
        Trigger(
            type="prometheus",
            metadata={
                "serverAddress": "http://prometheus-operated.default.svc.cluster.local",
                "metricName": "http_requests_total",
                "threshold": "100",
                "query": "sum(rate(http_requests_total{job=\"example-job\"}[2m]))"
            }
        )
    ]
    """,
    )

    isPublic: bool = Field(
        False,
        description="Whether the model group can be accessed through a public endpoint.",
    )


class ScalingConfig(PakaBaseModel):
    minInstances: int = Field(
        ..., description="The minimum number of instances to provision."
    )
    maxInstances: int = Field(
        ..., description="The maximum number of instances to provision."
    )

    @model_validator(mode="before")
    def check_instances_num(cls, values: Dict[str, int]) -> Dict[str, int]:
        min_instances, max_instances = values.get("minInstances"), values.get(
            "maxInstances"
        )
        if min_instances and max_instances and max_instances < min_instances:
            raise ValueError(
                "maxInstances must be greater than or equal to minInstances"
            )
        return values

    @field_validator("minInstances", mode="before")
    def validate_min_instances(cls, v: int) -> int:
        """
        Validates the value of the minInstances field. Spinning up 1 model group instance is heavy.
        No scaling down to 0 for now.

        Args:
            v (int): The value of the minInstances field.

        Returns:
            int: The input value if validation is successful.

        Raises:
            ValueError: If the value of the input value is invalid.
        """
        if v < 0:
            raise ValueError("minInstances cannot be negative")
        return v


class ScalingConfigNonZero(ScalingConfig):
    @field_validator("minInstances", mode="before")
    def validate_min_instances(cls, v: int) -> int:
        """
        Validates the value of the minInstances field. Spinning up 1 model group instance is heavy.
        No scaling down to 0 for now.

        Args:
            v (int): The value of the minInstances field.

        Returns:
            int: The input value if validation is successful.

        Raises:
            ValueError: If the value of the input value is invalid.
        """
        if v <= 0:
            raise ValueError("minInstances must be greater than 0")
        return v


class OnDemandModelGroup(CloudModelGroup, ScalingConfigNonZero):
    pass


class MixedModelGroup(CloudModelGroup):
    baseInstances: int = Field(
        ...,
        description=(
            "Initial set of instances that are provisioned using on-demand "
            "instances. These instances are always running. Their primary purpose is to guarantee the constant "
            "availability of the model group service. If you don't require these base instances, you can set their number to 0."
        ),
    )
    maxOnDemandInstances: int = Field(
        ...,
        description=(
            "This sets the maximum limit for provisioning on-demand instances, including the base instances. "
            "It should always be set to a value equal to or greater than 'baseInstances'. "
            "These on-demand instances serve as a fallback when spot instances are unavailable. "
            "The actual number of on-demand instances is dynamically adjusted by the autoscaler based on the workload requirements."
        ),
    )
    spot: ScalingConfigNonZero = Field(
        ...,
        description=(
            "This configuration sets the scaling parameters for spot instances within the node group. "
            "It specifies the minimum and maximum number of spot instances that can be used. "
            "Spot instances are cost-effective but may not always be available due to market conditions. "
            "The autoscaler uses this configuration to dynamically adjust the number of spot instances based on workload demand."
        ),
    )

    @model_validator(mode="before")
    def check_instances_num(cls, values: Dict[str, int]) -> Dict[str, int]:
        base_instances, max_on_demand_instances = values.get(
            "baseInstances"
        ), values.get("maxOnDemandInstances")
        if (
            base_instances is not None
            and max_on_demand_instances is not None
            and max_on_demand_instances < base_instances
        ):
            raise ValueError(
                "maxOnDemandInstances must be greater than or equal to baseInstances"
            )
        return values

    @field_validator("baseInstances", mode="before")
    def validate_base_instances(cls, v: int) -> int:
        if v < 0:
            raise ValueError("baseInstances must be greater than or equal to 0")
        return v


class AwsModelGroup(OnDemandModelGroup, AwsNode):
    pass


class AwsMixedModelGroup(MixedModelGroup, AwsNode):
    pass


class ClusterConfig(PakaBaseModel):
    """
    Represents the configuration for a cluster.
    """

    name: str = Field(..., description="The name of the cluster.")
    region: str = Field(..., description="The default region for the cluster.")
    namespace: str = Field(
        "default", description="The namespace in which the cluster is deployed."
    )
    nodeType: str = Field(..., description="The type of nodes in the cluster.")
    minNodes: int = Field(
        ..., description="The minimum number of nodes in the cluster."
    )
    maxNodes: int = Field(
        ..., description="The maximum number of nodes in the cluster."
    )
    logRetentionDays: int = Field(
        14, description="The number of days to retain log entries."
    )


class CloudVectorStore(CloudNode):
    """
    Represents a cloud vector store.
    """

    replicas: int = Field(1, description="The number of replicas for the vector store.")
    storageSize: str = Field(
        "10Gi", description="The size of the storage of one node for the vector store."
    )
    resourceRequest: Optional[ResourceRequest] = Field(
        None,
        description="The resource request for the vector store, specifying the amount of CPU and memory to request.",
    )

    @field_validator("storageSize", mode="before")
    def validate_storage_size(cls, v: str) -> str:
        """
        Validates the format of the storage_size field.

        Args:
            v (str): The value of the storage_size field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid size format")

    @field_validator("replicas", mode="before")
    def validate_replicas(cls, v: int) -> int:
        """
        Validates the value of the replicas field.

        Args:
            v (int): The value of the replicas field.

        Returns:
            int: The input value if validation is successful.

        Raises:
            ValueError: If the value of the input value is invalid.
        """
        if v <= 0:
            raise ValueError("replicas must be greater than 0")
        return v


class NodeGroup(ScalingConfigNonZero):
    nodeTypes: List[str] = Field(
        ..., description="The types of nodes in the node group."
    )
    diskSize: Optional[int] = Field(
        None, description="The size of the disk for the node in GB."
    )
    isSpot: bool = Field(..., description="Whether the node group uses spot instances.")


class Job(PakaBaseModel):
    """
    Represents a job cluster configuration.
    """

    enabled: bool = Field(False, description="Whether the job cluster is enabled.")
    brokerStorageSize: str = Field(
        "10Gi", description="The size of the storage for the broker."
    )

    nodeGroups: Optional[List[NodeGroup]] = Field(
        None,
        description="The node groups for job workers. If None, default node groups are used.",
    )

    @field_validator("brokerStorageSize", mode="before")
    def validate_broker_storage_size(cls, v: str) -> str:
        """
        Validates the format of the brokerStorageSize field.

        Args:
            v (str): The value of the brokerStorageSize field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid size format")


class Function(PakaBaseModel):
    """
    Configuration for function.
    """

    nodeGroups: Optional[List[NodeGroup]] = Field(
        None,
        description="The node groups for functions. If None, default node groups are used.",
    )


class Prometheus(PakaBaseModel):
    """
    Represents a Prometheus configuration.
    """

    enabled: bool = Field(False, description="Whether Prometheus is enabled.")
    storageSize: str = Field(
        "10Gi", description="The size of the storage for Prometheus."
    )
    grafana: bool = Field(False, description="Whether Grafana is enabled.")
    alertmanager: bool = Field(False, description="Whether Alertmanager is enabled.")
    kubeApiServer: bool = Field(
        False, description="Whether the Kubernetes API server is enabled."
    )
    kubelet: bool = Field(False, description="Whether the Kubelet is enabled.")
    kubeControllerManager: bool = Field(
        False, description="Whether the Kubernetes controller manager is enabled."
    )
    coreDns: bool = Field(False, description="Whether CoreDNS is enabled.")
    kubeEtcd: bool = Field(False, description="Whether the Kubernetes etcd is enabled.")
    kubeScheduler: bool = Field(
        False, description="Whether the Kubernetes scheduler is enabled."
    )
    kubeProxy: bool = Field(
        False, description="Whether the Kubernetes proxy is enabled."
    )
    kubeStateMetrics: bool = Field(
        False, description="Whether the Kubernetes state metrics are enabled."
    )
    nodeExporter: bool = Field(
        False, description="Whether the Node Exporter is enabled."
    )
    thanosRuler: bool = Field(False, description="Whether the Thanos Ruler is enabled.")

    @field_validator("storageSize", mode="before")
    def validate_storage_size(cls, v: str) -> str:
        """
        Validates the format of the storageSize field.

        Args:
            v (str): The value of the storageSize field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid size format")


class Tracing(PakaBaseModel):
    """
    Represents the configuration for tracing.
    """

    enabled: bool = Field(False, description="Whether tracing is enabled.")
    autoScalingEnabled: bool = Field(
        False, description="Whether auto-scaling is enabled for tracing."
    )
    zipkinHelmSettings: Optional[Dict[str, Any]] = Field(
        None,
        description="The settings for the Zipkin Helm chart. See https://github.com/openzipkin/zipkin-helm",
    )


T_OnDemandModelGroup = TypeVar("T_OnDemandModelGroup", bound=OnDemandModelGroup)
T_MixedModelGroup = TypeVar("T_MixedModelGroup", bound=MixedModelGroup)


class CloudConfig(PakaBaseModel, Generic[T_OnDemandModelGroup, T_MixedModelGroup]):
    """
    Represents the configuration for the cloud environment.
    """

    cluster: ClusterConfig = Field(
        ..., description="The configuration for the Kubernetes cluster."
    )
    modelGroups: Optional[List[T_OnDemandModelGroup]] = Field(
        None,
        description="The list of model groups to be deployed in the cloud. Default is None. If None, no model groups are deployed.",
    )
    mixedModelGroups: Optional[List[T_MixedModelGroup]] = Field(
        None,
        description="The list of mixed model groups to be deployed in the cloud. Default is None. If None, no mixed model groups are deployed.",
    )
    vectorStore: Optional[CloudVectorStore] = Field(
        None,
        description="The configuration for the vector store in the cloud. Default is None. If None, no vector store is deployed",
    )
    function: Optional[Function] = Field(
        None, description="The configuration for functions. Default is None."
    )
    job: Optional[Job] = Field(
        None,
        description="The configuration for the job broker. Default is None. If None, no job broker is deployed.",
    )
    prometheus: Optional[Prometheus] = Field(
        None,
        description="The configuration for Prometheus. Default is None. If None, Prometheus is not deployed.",
    )
    tracing: Optional[Tracing] = Field(
        None,
        description="The configuration for tracing. Default is None. If None, tracing is not enabled.",
    )

    @model_validator(mode="before")
    def check_model_group_names(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        # No model groups should have the same name
        model_group_names = set()

        for group in values.get("modelGroups", []):
            group_name = dict(group)["name"]
            if group_name in model_group_names:
                raise ValueError(f"Duplicate model group names are not allowed")
            model_group_names.add(group_name)

        for mixed_group in values.get("mixedModelGroups", []):
            mixed_group_name = dict(mixed_group)["name"]
            if mixed_group_name in model_group_names:
                raise ValueError(f"Duplicate model group names are not allowed")
            model_group_names.add(mixed_group_name)

        return values


class AwsConfig(CloudConfig[AwsModelGroup, AwsMixedModelGroup]):
    modelGroups: Optional[List[AwsModelGroup]] = Field(
        None,
        description="The list of model groups to be deployed in the cloud. Default is None. If None, no model groups are deployed.",
    )

    mixedModelGroups: Optional[List[AwsMixedModelGroup]] = Field(
        None,
        description="The list of mixed model groups to be deployed in the cloud. Default is None. If None, no mixed model groups are deployed.",
    )


class Config(PakaBaseModel):
    """
    Configuration class for managing cloud cluster settings.
    """

    version: str = Field(..., description="The version of the configuration.")
    aws: Optional[AwsConfig] = Field(None, description="The AWS cloud configuration.")

    @model_validator(mode="before")
    def check_one_field(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validates that exactly one cloud configuration is provided.

        Args:
            values (Dict[str, Any]): Dictionary of field values for the Config class.

        Returns:
            Dict[str, Any]: The input values if validation is successful.

        Raises:
            ValueError: If more or less than one cloud configuration is provided.
        """
        if "aws" not in values:
            raise ValueError("Exactly one cloud configuration must be provided")

        return values

    @field_validator("version", mode="before")
    def validate_version(cls, v: str) -> str:
        if not re.match(r"^\d+\.\d+$", v):
            raise ValueError('version must be in the format "x.x"')
        return v


def generate_yaml(config: Config) -> str:
    """
    Generate a YAML string representation of the given config object.

    Args:
        config (Config): The config object to generate YAML from.

    Returns:
        str: The YAML string representation of the config object.
    """
    return to_yaml(config.model_dump(exclude_none=True))


def parse_yaml(yaml_str: str) -> Config:
    """
    Parse a YAML string and return a Config object.

    Args:
        yaml_str (str): The YAML string to parse.

    Returns:
        Config: The parsed Config object.
    """
    yaml = YAML()
    data = yaml.load(yaml_str)
    version = data.get("version", None)
    if version is None:
        raise ValueError("Invalid configuration: The 'version' field is missing.")

    if not re.match(r"^\d+\.\d+$", version):
        raise ValueError('version must be in the format "x.x"')

    # Make sure the major version matches
    major_version, minor_version = map(int, version.split("."))
    tool_major_version, tool_minor_version = map(int, CONFIG_VERSION.split("."))

    if major_version < tool_major_version:
        raise ValueError(
            f"Invalid configuration: This tool supports versions starting from {tool_major_version}.0."
            " Please use an older version of the tool if you need to work with a previous configuration version."
        )
    elif major_version > tool_major_version:
        raise ValueError(
            f"Invalid configuration: Your current tool is too old. Please upgrade your tool to handle this configuration."
        )
    elif minor_version > tool_minor_version:  # No forward compatibility
        raise ValueError(
            f"Invalid configuration: This tool supports versions up to {tool_major_version}.{tool_minor_version}."
            " Please upgrade your tool to handle this configuration."
        )

    return Config(**data)



from typing import Any

from paka.cluster.context import Context
from paka.model.store import ModelStore, S3ModelStore


def get_model_store(ctx: Context, *args: Any, **kwargs: Any) -> ModelStore:
    assert ctx.provider == "aws"

    return S3ModelStore(ctx.bucket, *args, **kwargs)



from __future__ import annotations

from typing import Optional

import fasteners
import pulumi_kubernetes as k8s

from paka.config import CloudConfig, Config


class Context:
    _k8s_provider: Optional[k8s.Provider]
    _config: Optional[Config]
    # Materialized bucket with a unique name
    _bucket: Optional[str]
    # Materialized container registry url
    _registry: Optional[str]
    # The kubeconfig str
    _kubeconfig: Optional[str]

    # Need to lock the access to these fields
    _should_save_kubeconfig: bool = False

    def __init__(self) -> None:
        # Ugly, ideally, we can create these locks dynamically in __getattr__.
        # However, __getattr__ is not thread safe either. We need another lock to protect the creation of locks.
        # This lock is going to be a bottleneck. Therefore, we pre-create the locks.
        # Multiple locks pose a risk of deadlock. We need to be careful when acquiring multiple locks.
        self._k8s_provider_lock = fasteners.ReaderWriterLock()
        self._config_lock = fasteners.ReaderWriterLock()
        self._bucket_lock = fasteners.ReaderWriterLock()
        self._registry_lock = fasteners.ReaderWriterLock()
        self._kubeconfig_lock = fasteners.ReaderWriterLock()

    @fasteners.write_locked(lock="_k8s_provider_lock")
    def set_k8s_provider(self, k8s_provider: k8s.Provider) -> None:
        self._k8s_provider = k8s_provider

    @property
    @fasteners.read_locked(lock="_k8s_provider_lock")
    def k8s_provider(self) -> Optional[k8s.Provider]:
        return self._k8s_provider

    @fasteners.write_locked(lock="_config_lock")
    def set_config(self, config: Config) -> None:
        self._config = config

    @property
    @fasteners.read_locked(lock="_config_lock")
    def config(self) -> Optional[Config]:
        return self._config

    @property
    @fasteners.read_locked(lock="_config_lock")
    def cloud_config(self) -> Optional[CloudConfig]:
        if self._config is None:
            raise RuntimeError("Config is not set.")
        if self._config.aws is None:
            raise RuntimeError("Only AWS is supported.")

        return self._config.aws

    @property
    @fasteners.read_locked(lock="_config_lock")
    def region(self) -> Optional[str]:
        # fasteners's inter thread reader lock is reentrant. We can call other methods that acquire the same lock.
        # https://github.com/harlowja/fasteners/blob/06c3f06cab4e135b8d921932019a231c180eb9f4/docs/guide/inter_thread.md#lack-of-features
        return self.cloud_config.cluster.region

    @property
    @fasteners.read_locked(lock="_config_lock")
    def namespace(self) -> Optional[str]:
        # reentrant
        return self.cloud_config.cluster.namespace

    @property
    @fasteners.read_locked(lock="_config_lock")
    def provider(self) -> str:
        # reentrant
        _ = self.cloud_config
        return "aws"

    @property
    @fasteners.read_locked(lock="_config_lock")
    def cluster_name(self) -> str:
        # reentrant
        return self.cloud_config.cluster.name

    @fasteners.write_locked(lock="_bucket_lock")
    def set_bucket(self, bucket: str) -> None:
        self._bucket = bucket

    @property
    @fasteners.read_locked(lock="_bucket_lock")
    def bucket(self) -> Optional[str]:
        return self._bucket

    @fasteners.write_locked(lock="_registry_lock")
    def set_registry(self, registry: str) -> None:
        self._registry = registry

    @property
    @fasteners.read_locked(lock="_registry_lock")
    def registry(self) -> Optional[str]:
        return self._registry

    @fasteners.write_locked(lock="_kubeconfig_lock")
    def set_kubeconfig(self, kubeconfig: str) -> None:
        self._kubeconfig = kubeconfig

    @property
    @fasteners.read_locked(lock="_kubeconfig_lock")
    def kubeconfig(self) -> Optional[str]:
        return self._kubeconfig

    def set_should_save_kubeconfig(self, should_save_kubeconfig: bool) -> None:
        self._should_save_kubeconfig = should_save_kubeconfig

    @property
    def should_save_kubeconfig(self) -> bool:
        return self._should_save_kubeconfig



from __future__ import annotations

from typing import Sequence

import pulumi
import pulumi_aws as aws
import pulumi_eks as eks
from pulumi import Input

from paka.cluster.context import Context
from paka.utils import get_instance_info


def odic_role_for_sa(
    ctx: Context,
    cluster: eks.Cluster,
    role_name: str,
    ns_service_account: str,
) -> aws.iam.Role:
    """
    Creates an IAM role for a service account in an EKS cluster using OpenID Connect (OIDC) authentication.

    Args:
        config (CloudConfig): The cloud configuration.
        cluster (eks.Cluster): The EKS cluster.
        role_name (str): The name of the role.
        ns_service_account (str): The name of the service account. e.g. "default:sa", "kube-system:auto-scaler"

    Returns:
        aws.iam.Role: The IAM role for the service account.
    """
    oidc_url = cluster.core.oidc_provider.url
    oidc_arn = cluster.core.oidc_provider.arn

    assume_role_policy = pulumi.Output.all(oidc_url, oidc_arn).apply(
        lambda args: aws.iam.get_policy_document(
            statements=[
                aws.iam.GetPolicyDocumentStatementArgs(
                    effect="Allow",
                    principals=[
                        aws.iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Federated",
                            identifiers=[str(args[1])],
                        )
                    ],
                    actions=["sts:AssumeRoleWithWebIdentity"],
                    conditions=[
                        aws.iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable=f"{args[0]}:sub",
                            values=[f"system:serviceaccount:{ns_service_account}"],
                        )
                    ],
                )
            ],
        ).json
    )

    role = aws.iam.Role(
        f"{ctx.cluster_name}-{role_name}-role",
        assume_role_policy=assume_role_policy,
    )

    return role


def get_ami_for_instance(ctx: Context, instance_type: str) -> str:
    instance_info = get_instance_info(ctx.provider, ctx.region, instance_type)
    gpu_count = instance_info.get("gpu_count", 0) or 0
    arch = instance_info.get("arch", "x86_64")

    if gpu_count > 0:
        if arch == "x86_64":
            return "AL2_x86_64_GPU"
        else:
            return "BOTTLEROCKET_ARM_64_NVIDIA"
    else:
        if arch == "arm64":
            return "AL2_ARM_64"
    return "AL2_x86_64"


def create_vpc_endpoint_for_s3(
    vpc_id: str, route_table_ids: Input[Sequence[Input[str]]], region: str
) -> aws.ec2.VpcEndpoint:
    s3_service_name = f"com.amazonaws.{region}.s3"

    vpc_endpoint = aws.ec2.VpcEndpoint(
        "s3-vpc-endpoint",
        vpc_id=vpc_id,
        service_name=s3_service_name,
        route_table_ids=route_table_ids,
    )
    return vpc_endpoint



from __future__ import annotations

import json
import os
from typing import List, Optional

import typer
from kubernetes import client

from paka.cli.utils import (
    get_cluster_namespace,
    load_kubeconfig,
    process_envs,
    resolve_image,
)
from paka.k8s.job.worker import create_workers, delete_workers
from paka.logger import logger
from paka.utils import kubify_name

job_app = typer.Typer()


def prefixed_job_name(job_name: str) -> str:
    # Prefix the job name with "job-" to reduce the chances of name collision
    # between jobs and functions.
    if not job_name.startswith("job-"):
        return f"job-{job_name}"
    return job_name


@job_app.command()
def deploy(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
    name: str = typer.Option(
        "",
        "--name",
        help="The name of the job. This name will be used to track the job's "
        "progress and manage its resources. If not provided, the system will use "
        "the base name of the 'source_dir' or the 'image' name as the job name.",
    ),
    entrypoint: str = typer.Option(
        ...,
        "--entrypoint",
        help="The entrypoint of the application. This refers to the command "
        "defined in the Procfile that will be executed to start the application "
        "when the job is run.",
    ),
    source_dir: Optional[str] = typer.Option(
        None,
        "--source",
        help="The directory containing the source code of the application. If "
        "specified, a new Docker image will be built using the source code from "
        "this directory. A Dockerfile is not required because the build process "
        "uses Cloud Native's Buildpacks, which automatically detect and install "
        "dependencies.",
    ),
    image: Optional[str] = typer.Option(
        None,
        "--image",
        help="The name of the Docker image to deploy. If both an image and a "
        "source directory are provided, this image will be used and the source "
        "directory will be ignored.",
    ),
    max_workers: int = typer.Option(
        5,
        "--max-workers",
        help="The maximum number of workers that can be allocated for this job. "
        "This sets an upper limit on the number of workers that the autoscaling "
        "system can create to handle tasks for this job.",
    ),
    tasks_per_worker: int = typer.Option(
        5,
        "--tasks-per-worker",
        help="The number of tasks that need to be allocated to each worker before "
        "the autoscaler creates a new worker.",
    ),
    wait_existing_tasks: bool = typer.Option(
        True,
        "--wait-existing-tasks",
        help="Determines whether the system should wait for existing tasks to "
        "complete before deploying the new job. If set to true, the deployment "
        "will wait until all current tasks have finished.",
    ),
    env_vars: List[str] = typer.Option(
        [],
        "--env",
        help="Specify environment variables for the function in the format 'key=value'",
        show_default=False,
    ),
    resource_requests: Optional[str] = typer.Option(
        None,
        "--resource-requests",
        help='The resource requests for the function, in JSON format. For example: \'{"cpu": "100m", "memory": "128Mi"}\'',
    ),
    resource_limits: Optional[str] = typer.Option(
        None,
        "--resource-limits",
        help='The resource limits for the function, in JSON format. For example: \'{"cpu": "200m", "memory": "256Mi"}\'',
    ),
) -> None:
    """
    Deploy a job.
    """
    load_kubeconfig(cluster_name)
    resolved_image = resolve_image(cluster_name, image, source_dir)

    if image:
        job_name = image
    elif source_dir:
        job_name = os.path.basename(source_dir)

    envs = process_envs(env_vars)
    resource_requests_dict = (
        json.loads(resource_requests) if resource_requests else None
    )
    resource_limits_dict = json.loads(resource_limits) if resource_limits else None

    create_workers(
        namespace=get_cluster_namespace(cluster_name),
        job_name=kubify_name(prefixed_job_name(name or job_name)),
        image=resolved_image,
        entrypoint=entrypoint,
        tasks_per_worker=tasks_per_worker,
        max_replicas=max_workers,
        drain_existing_job=wait_existing_tasks,
        envs=envs,
        resource_requests=resource_requests_dict,
        resource_limits=resource_limits_dict,
    )


@job_app.command()
def delete(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
    name: str = typer.Argument(
        ...,
        help="The name of the job to delete.",
    ),
    wait_existing_tasks: bool = typer.Option(
        True,
        "--wait-existing-tasks",
        help="Determines whether the system should wait for existing tasks to "
        "complete before deploying the new job. If set to true, the deployment "
        "will wait until all current tasks have finished.",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Automatic yes to prompts. Use this option to bypass the confirmation "
        "prompt and directly proceed with the deletion.",
    ),
) -> None:
    """
    Delete a job.
    """
    if yes or typer.confirm(
        f"Are you sure you want to delete the job {name}?", default=False
    ):
        load_kubeconfig(cluster_name)
        logger.info(f"Deleting job {name}")
        delete_workers(
            get_cluster_namespace(cluster_name),
            prefixed_job_name(name),
            wait_existing_tasks,
        )
        logger.info(f"Successfully deleted job {name}")


@job_app.command()
def list(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
) -> None:
    """
    Lists all jobs.
    """
    load_kubeconfig(cluster_name)
    api_instance = client.AppsV1Api()

    label_selector = "role=worker"

    # List the deployments in the specified namespace that match the field selector
    deployments = api_instance.list_namespaced_deployment(
        namespace=get_cluster_namespace(cluster_name), label_selector=label_selector
    )

    for deployment in deployments.items:
        if deployment.metadata and deployment.metadata.name:
            logger.info(deployment.metadata.name)

    if not deployments.items:
        logger.info("No jobs found.")



from __future__ import annotations

import hashlib
import json
import os
import random
import re
import string
import tempfile
from contextlib import contextmanager
from enum import Enum
from functools import lru_cache, wraps
from io import StringIO
from pathlib import Path
from typing import Any, Callable, Dict, Generator, Optional, TypeVar, cast

import boto3
import requests
from ruamel.yaml import YAML

from paka.constants import HOME_ENV_VAR, PROJECT_NAME, PULUMI_STACK_NAME

T = TypeVar("T", bound=Callable[..., Any])


def camel_to_kebab(camel_str: str) -> str:
    """
    Converts a camel case string to kebab case.

    Args:
        camel_str (str): The camel case string to be converted.

    Returns:
        str: The kebab case string.

    Example:
        >>> camel_to_kebab("camelCaseString")
        'camel-case-string'
    """
    camel_str = re.sub("(.)([A-Z][a-z]+)", r"\1-\2", camel_str)
    return re.sub("([a-z0-9])([A-Z])", r"\1-\2", camel_str).lower()


def kubify_name(old: str) -> str:
    """
    Convert a string into a valid Kubernetes name.

    This function takes a string, converts it to lowercase, replaces disallowed characters with '-',
    trims leading non-alphabetic characters, trims trailing non-alphanumeric characters, and truncates
    it to a maximum length of 63 characters to create a valid Kubernetes name.

    If the resulting name is empty, it raises an exception.

    Args:
        old (str): The original string to be converted.

    Returns:
        str: The converted string that is a valid Kubernetes name.

    Raises:
        Exception: If the resulting name is empty.
    """
    max_len = 63

    new_name = old.lower()

    # replace disallowed chars with '-'
    new_name = re.sub(r"[^-a-z0-9]", "-", new_name)

    # trim leading non-alphabetic
    new_name = re.sub(r"^[^a-z]+", "", new_name)

    # trim trailing
    new_name = re.sub(r"[^a-z0-9]+$", "", new_name)

    # truncate to length
    if len(new_name) > max_len:
        new_name = new_name[:max_len]

    if len(new_name) == 0:
        raise Exception(f"Name: {old} can't be converted to a valid Kubernetes name")

    return new_name


def get_project_data_dir() -> str:
    """
    Get the project data directory.

    This function retrieves the directory path where the project data is stored.
    If the environment variable HOME_ENV_VAR is set, it returns its value.
    Otherwise, it returns the home directory appended with the kebab-case project name.

    Returns:
        str: The absolute path of the project data directory.
    """
    return os.environ.get(
        HOME_ENV_VAR, str(Path.home() / f".{camel_to_kebab(PROJECT_NAME)}")
    )


def call_once(func: T) -> T:
    """
    Decorator to ensure a function is only executed once.
    """
    has_been_called = False

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        nonlocal has_been_called
        if not has_been_called:
            has_been_called = True
            return func(*args, **kwargs)

    return cast(T, wrapper)


def to_yaml(obj: Dict[Any, Any]) -> str:
    """
    Converts an dictionary to a YAML string.

    Args:
        obj (dict): The dictionary to be converted.

    Returns:
        str: The YAML string.
    """
    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)
    buf = StringIO()
    yaml.dump(obj, buf)
    return buf.getvalue()


def read_yaml_file(path: str) -> Dict[str, Any]:
    """
    Reads a YAML file and returns its contents as a dictionary.

    This function opens a YAML file, reads its contents, and converts it into a Python dictionary.
    If the file does not exist, it returns an empty dictionary.

    Args:
        path (str): The path to the YAML file.

    Returns:
        Dict[str, Any]: The contents of the YAML file as a dictionary, or an empty dictionary if the file does not exist.
    """
    yaml = YAML()
    try:
        with open(path, "r") as file:
            data = yaml.load(file)
    except FileNotFoundError:
        data = {}
    return data or {}


def get_cluster_data_dir(cluster_name: str) -> str:
    """
    Get the cluster data directory.

    Args:
        name (str): The name of the cluster.

    Returns:
        str: The cluster data directory.
    """
    return os.path.join(get_project_data_dir(), "clusters", cluster_name)


def get_pulumi_root() -> str:
    """
    Get the pulumi data directory.

    Returns:
        str: The pulumi data directory.
    """
    return str(Path(get_project_data_dir()) / "pulumi")


def save_kubeconfig(cluster_name: str, kubeconfig_json: Optional[str]) -> None:
    """
    Save the kubeconfig data as a YAML file named 'kubeconfig.yaml'.

    This function takes the kubeconfig data in JSON format, converts it to YAML,
    and saves it to a file named 'kubeconfig.yaml'. The file is stored
    in the directory specified by the cluster name.

    Args:
        cluster_name (str): The name of the cluster. This is used to determine the directory where the kubeconfig file will be saved.
        kubeconfig_json (str): The kubeconfig data in JSON format.

    Returns:
        None
    """
    if kubeconfig_json is None:
        return

    kubeconfig_data = json.loads(kubeconfig_json)

    kubeconfig_file_path = os.path.join(
        get_cluster_data_dir(cluster_name), "kubeconfig.yaml"
    )
    os.makedirs(os.path.dirname(kubeconfig_file_path), exist_ok=True)

    with open(kubeconfig_file_path, "w") as f:
        f.write(to_yaml(kubeconfig_data))


def random_str(length: int = 5) -> str:
    """
    Generate a random string of a specified length.

    This function generates a random string of a given length. The string consists of
    both ASCII letters (both lowercase and uppercase) and digits.

    Args:
        length (int, optional): The length of the random string to be generated. Defaults to 5.

    Returns:
        str: The generated random string.
    """
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def calculate_sha256(file_path: str) -> str:
    """
    Calculate the SHA-256 hash of a file.

    Args:
        file_path (str): The path to the file for which the SHA-256 hash is to be calculated.

    Returns:
        str: The SHA-256 hash of the file, as a hexadecimal string.
    """
    sha256_hash = hashlib.sha256()

    with open(file_path, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()


def get_gh_release_latest_version(repo: str) -> str:
    """
    Get the latest release version of a GitHub repository.

    This function queries the GitHub API to get the latest release version of a repository.

    Args:
        repo (str): The GitHub repository in the format 'owner/repo'.

    Returns:
        str: The latest release version of the repository.
    """
    url = f"https://api.github.com/repos/{repo}/releases/latest"
    response = requests.get(url)
    response.raise_for_status()
    data = response.json()
    return data["tag_name"]


@contextmanager
def download_url(url: str) -> Generator[str, None, None]:
    """
    Download a file from a URL and return the path to the downloaded file.

    Args:
        url (str): The URL of the file to be downloaded.

    Returns:
        str: The path to the downloaded file.
    """
    fd, tmp_file = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "wb") as tf:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                for chunk in r.iter_content(chunk_size=8192):
                    tf.write(chunk)

            tf.flush()
            os.fsync(tf.fileno())

        yield tmp_file
    finally:
        os.remove(tmp_file)


class PulumiStackKey(Enum):
    NAMESPACE = "namespace"
    REGION = "region"
    PROVIDER = "provider"
    REGISTRY = "registry"
    BUCKET = "bucket"
    KUBECONFIG = "kubeconfig"


def read_pulumi_stack(cluster_name: str, key: str) -> Any:
    try:
        k = PulumiStackKey[key.upper()]
    except KeyError:
        raise ValueError(
            f"Invalid key: {key}. Expected one of: {list(PulumiStackKey.__members__.values())}"
        )
    stack_json = _load_pulumi_stack(cluster_name)

    return _read_pulumi_stack_by_key(stack_json, k)


@lru_cache(maxsize=100)
def _load_pulumi_stack(cluster_name: str) -> dict:
    pulumi_backend_url = os.environ.get("PULUMI_BACKEND_URL", "")

    if not pulumi_backend_url:
        raise Exception("Pulumi backend URL is not set")

    if pulumi_backend_url.startswith("file://"):
        pulumi_root = Path(get_pulumi_root())  # Windows support?
        pulumi_stack_file = (
            pulumi_root
            / ".pulumi"
            / "stacks"
            / cluster_name
            / f"{PULUMI_STACK_NAME}.json"
        )
        stack_json = json.loads(pulumi_stack_file.read_text())
    elif pulumi_backend_url.startswith("s3://"):
        s3 = boto3.client("s3")
        response = s3.get_object(
            Bucket=pulumi_backend_url[5:],
            Key=f".pulumi/stacks/{cluster_name}/{PULUMI_STACK_NAME}.json",
        )
        stack_json = json.loads(response["Body"].read().decode("utf-8"))
    else:
        raise Exception("Unsupported Pulumi backend URL")

    return stack_json


def _read_pulumi_stack_by_key(stack_json: dict, k: PulumiStackKey) -> Any:
    resources = stack_json["checkpoint"]["latest"]["resources"]

    # For now, we assume that the cloud is AWS
    if k == PulumiStackKey.PROVIDER:
        for resource in resources:
            if resource["type"] == "pulumi:providers:aws":
                return "aws"
    elif k == PulumiStackKey.REGION:
        for resource in resources:
            if resource["type"] == "pulumi:providers:aws":
                return resource["outputs"]["region"]
    elif k == PulumiStackKey.REGISTRY:
        for resource in resources:
            # Since we creating only one ECR repository, there is no need to use the resource urn
            if resource["type"] == "aws:ecr/repository:Repository":
                return resource["outputs"]["repositoryUrl"]
    elif k == PulumiStackKey.BUCKET:
        for resource in resources:
            # Since we creating only one S3 bucket, there is no need to use the resource urn
            if resource["type"] == "aws:s3/bucket:Bucket":
                return resource["outputs"]["bucket"]
    elif k == PulumiStackKey.NAMESPACE:
        for resource in resources:
            if resource["type"] == "kubernetes:core/v1:Namespace" and resource[
                "urn"
            ].endswith("app-ns"):
                return resource["outputs"]["metadata"]["name"]
        # If no namespace is found, return the default namespace
        return "default"
    elif k == PulumiStackKey.KUBECONFIG:
        for resource in resources:
            # Since we creating only one secret, there is no need to use the resource urn
            if resource["type"] == "eks:index:Cluster":
                return resource["outputs"]["core"]["kubeconfig"]

    raise Exception(f"Unsupported PulumiStackKey: {k}")


@lru_cache(maxsize=100)
def get_instance_info(provider: str, region: str, instance_type: str) -> Dict[str, Any]:
    if provider == "aws":
        # Get the CPU, memory and GPU count for the instance type with boto3
        ec2 = boto3.client("ec2", region_name=region)

        response = ec2.describe_instance_types(InstanceTypes=[instance_type])  # type: ignore

        instance_types = response.get("InstanceTypes", [])
        if instance_types:
            instance_type_info = instance_types[0]
            gpu_info = instance_type_info.get("GpuInfo", {})
            gpu, vram = gpu_info.get("Gpus", [{}])[0], gpu_info.get(
                "TotalGpuMemoryInMiB", 0
            )
            architectures = instance_type_info.get("ProcessorInfo", {}).get(
                "SupportedArchitectures", []
            )
            arch = architectures[0] if architectures else None
            return {
                "cpu": instance_type_info.get("VCpuInfo", {}).get("DefaultVCpus"),
                "memory": instance_type_info.get("MemoryInfo", {}).get("SizeInMiB"),
                "gpu_count": gpu.get("Count", 0),
                "vram": vram,
                "gpu_manufacturer": gpu.get("Manufacturer"),
                "gpu_name": gpu.get("Name"),
                "arch": arch,
            }
    else:
        raise Exception(f"Unsupported provider: {provider}")
    raise Exception(f"Unsupported provider: {provider}")


def camel_to_snake(camel_str: str) -> str:
    kebab_case = camel_to_kebab(camel_str)
    return kebab_case.replace("-", "_")



"""
v1 supports mixed on-demand and spot instances. Some key design decisions:
a) Fail safe on-demand pods and instances. A safe number of on-demand instances are always maintained.
b) Use spot instances for cost savings. Spot instances are preferred over on-demand instances. If the spot instance
   pool is exhausted, on-demand instances are used.
c) Once the spot instance pool has capacity, the on-demand pods (except the fail safe pods) are rescheduled to spot instances.
"""

from __future__ import annotations

import json

from kubernetes import client
from kubernetes import config as k8s_config
from kubernetes.client.exceptions import ApiException

from paka.cluster.context import Context
from paka.cluster.utils import get_model_store
from paka.config import T_MixedModelGroup
from paka.k8s.model_group.ingress import create_model_vservice
from paka.k8s.model_group.service import (
    create_pod,
    create_scaled_object,
    create_service,
    create_service_monitor,
)
from paka.k8s.utils import apply_resource
from paka.logger import logger
from paka.model.hf_model import HuggingFaceModel
from paka.utils import kubify_name


def ensure_priority_class(name: str, priority: int) -> None:
    """
    Ensure that the priority class exists in the cluster.
    """
    name = kubify_name(name)
    api_instance = client.SchedulingV1Api()

    priority_class = client.V1PriorityClass(
        api_version="scheduling.k8s.io/v1",
        kind="PriorityClass",
        metadata=client.V1ObjectMeta(name=name),
        value=priority,
    )
    try:
        api_instance.read_priority_class(name)
        api_instance.replace_priority_class(name, body=priority_class)

    except ApiException as e:
        if e.status == 404:
            api_instance.create_priority_class(body=priority_class)
        else:
            raise e


def ensure_pdb(namespace: str, model_group: T_MixedModelGroup) -> None:
    """
    Ensure that the PodDisruptionBudget exists for the model group.
    """
    pdb = client.V1PodDisruptionBudget(
        kind="PodDisruptionBudget",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
        ),
        spec=client.V1PodDisruptionBudgetSpec(
            # Will slow down the speed of scaling down pods. But spot instances are usually terminated with 2 minutes.
            max_unavailable="30%",
            selector=client.V1LabelSelector(
                match_labels={
                    "app": "model-group",
                    "model": model_group.name,
                }
            ),
        ),
    )

    policy_v1 = client.PolicyV1Api()

    assert pdb.metadata and pdb.metadata.name and pdb.metadata.namespace

    try:
        existing_pdb = policy_v1.read_namespaced_pod_disruption_budget(
            name=pdb.metadata.name, namespace=pdb.metadata.namespace
        )
        assert existing_pdb.metadata
        pdb.metadata.resource_version = existing_pdb.metadata.resource_version
        policy_v1.replace_namespaced_pod_disruption_budget(
            name=pdb.metadata.name, namespace=pdb.metadata.namespace, body=pdb
        )
    except ApiException as e:
        if e.status == 404:
            policy_v1.create_namespaced_pod_disruption_budget(
                namespace=pdb.metadata.namespace, body=pdb
            )
        else:
            raise


def create_fail_safe_deployment(
    namespace: str, model_group: T_MixedModelGroup, pod: client.V1PodTemplateSpec
) -> client.V1Deployment:

    # Lack of optional chaining support in Python makes this code a bit verbose.
    assert (
        pod.spec
        and pod.spec.affinity
        and pod.spec.affinity.node_affinity
        and pod.spec.affinity.node_affinity.required_during_scheduling_ignored_during_execution
    )

    # We need to change the pod's affinity to ensure that it is scheduled on an on-demand instance.
    pod.spec.affinity.node_affinity.required_during_scheduling_ignored_during_execution.node_selector_terms.append(
        client.V1NodeSelectorTerm(
            match_expressions=[
                client.V1NodeSelectorRequirement(
                    key="lifecycle",
                    operator="In",
                    values=["on-demand"],
                )
            ]
        )
    )

    # We would also like to ensure that the pod is scheduled on a node with a higher priority.
    # This would reduce the chances of the pod being preempted by keda or HPA.
    # CA will also try to fullfil the pods with higher priority first.
    priority_class = "fail-safe"
    ensure_priority_class(priority_class, 100000)
    pod.spec.priority_class_name = priority_class

    return client.V1Deployment(
        api_version="apps/v1",
        kind="Deployment",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}-baseline",
            namespace=namespace,
        ),
        spec=client.V1DeploymentSpec(
            replicas=model_group.baseInstances,
            selector=client.V1LabelSelector(
                match_labels={
                    "app": "model-group",
                    "model": model_group.name,
                }
            ),
            template=pod,
        ),
    )


def create_auto_scale_deployment(
    namespace: str, model_group: T_MixedModelGroup, pod: client.V1PodTemplateSpec
) -> client.V1Deployment:

    ensure_pdb(namespace, model_group)

    # We want to change the pod's affinity to ensure that it prefers to be scheduled on a spot instances.
    # With this change, the pod will be scheduled on a spot instance if available.
    # However, if no instances are available, CA doesn't respect below preferred affinity as it is not a required affinity.
    # In such cases, CA relies on the priority expander to scale out more instances. We have given a higher priority to spot instances.
    assert pod.spec and pod.spec.affinity and pod.spec.affinity.node_affinity
    pod.spec.affinity.node_affinity.preferred_during_scheduling_ignored_during_execution = [
        client.V1PreferredSchedulingTerm(
            preference=client.V1NodeSelectorTerm(
                match_expressions=[
                    client.V1NodeSelectorRequirement(
                        key="lifecycle",
                        operator="In",
                        values=["spot"],
                    )
                ]
            ),
            weight=100,
        )
    ]

    return client.V1Deployment(
        api_version="apps/v1",
        kind="Deployment",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
        ),
        spec=client.V1DeploymentSpec(
            replicas=model_group.spot.minInstances,  # Scaler will update this
            selector=client.V1LabelSelector(
                match_labels={
                    "app": "model-group",
                    "model": model_group.name,
                }
            ),
            template=pod,
        ),
    )


def create_model_group_service(
    ctx: Context,
    namespace: str,
    model_group: T_MixedModelGroup,
) -> None:
    """
    Creates a Kubernetes service for a machine learning model group.

    Args:
        namespace (str): The namespace to create the service in.
        config (Config): The configuration for the service.
        model_group (T_CloudModelGroup): The model group to create the service for.

    Raises:
        ValueError: If the AWS configuration is not provided.

    Returns:
        None
    """
    k8s_config.load_kube_config_from_dict(json.loads(ctx.kubeconfig))

    config = ctx.cloud_config
    # Download the model to S3 first
    if model_group.model and model_group.model.useModelStore:
        if model_group.model.hfRepoId:
            model = HuggingFaceModel(
                name=model_group.name,
                repo_id=model_group.model.hfRepoId,
                files=model_group.model.files,
                model_store=get_model_store(ctx),
            )
            # If the model is not already in the model store, save it
            # That means users cannot update the model in the model store
            # They have to create a new model group or delete the old one
            if not model.model_store.glob(f"{model_group.name}/*"):
                model.save()
            else:
                logger.info(
                    f"Model {model_group.name} already exists in the model store. Skipping download."
                )

    port = 8000

    pod = create_pod(
        ctx,
        namespace,
        model_group,
        port,
    )

    fail_safe_deployment = create_fail_safe_deployment(namespace, model_group, pod)
    apply_resource(fail_safe_deployment)

    auto_scale_deployment = create_auto_scale_deployment(namespace, model_group, pod)
    apply_resource(auto_scale_deployment)

    # Service will direct traffic to pods managed by both deployments
    svc = create_service(namespace, model_group, port)
    apply_resource(svc)

    # Prometheus will monitor pods managed by both deployments
    if config.prometheus and config.prometheus.enabled:
        create_service_monitor(namespace, model_group)

    # Horizontal pod autoscaler will only scale the auto_scale_deployment, fail_safe_deployment is not scaled
    scaled_object = create_scaled_object(
        namespace,
        model_group,
        auto_scale_deployment,
        model_group.spot.minInstances,
        # Might result in many pending pods if spot.maxInstances > maxOnDemandInstances when spot pool is not available
        max(
            model_group.maxOnDemandInstances,
            model_group.spot.maxInstances,
        ),
    )
    if scaled_object:
        apply_resource(scaled_object)

    # Create a vservice to export the model group to the outside world
    if model_group.isPublic:
        create_model_vservice(namespace, model_group.name)



from __future__ import annotations

import re
import shlex
from typing import Any, Dict, List, Literal, Optional, Tuple

from kubernetes import client
from kubernetes.client.exceptions import ApiException
from kubernetes.dynamic import DynamicClient  # type: ignore

from paka.k8s.utils import apply_resource

VALID_RESOURCES = ["cpu", "memory"]

VALID_RESOURCES_GPU = ["nvidia.com/gpu"]


def validate_resource(resource: str, value: str) -> None:
    if resource == "cpu":
        if not re.match(r"^\d+(m)?$", value):
            raise ValueError("Invalid CPU value")
    elif resource == "memory":
        if not re.match(r"^\d+(Mi|Gi)$", value):
            raise ValueError("Invalid memory value")
    elif resource == "nvidia.com/gpu":
        if not value.isdigit() or int(value) < 1:
            raise ValueError("Invalid GPU value")


def enable_scale_to_zero(namespace: str = "knative-serving") -> None:
    """
    Enable scale to zero for the knative-serving namespace.
    """
    config_map = client.V1ConfigMap(
        api_version="v1",
        kind="ConfigMap",
        metadata=client.V1ObjectMeta(
            name="config-autoscaler",
            namespace=namespace,
        ),
        data={
            "enable-scale-to-zero": "true",
        },
    )

    api_instance = client.CoreV1Api()
    api_instance.patch_namespaced_config_map(
        name="config-autoscaler",
        namespace="knative-serving",
        body=config_map,
    )


def create_knative_service(
    service_name: str,
    namespace: str,
    image: str,
    entrypoint: str,
    min_instances: int,
    max_instances: int,
    # concurrency refers to the number of simultaneous requests that a single pod can handle
    # rps refers to requests per second a single pod can handle
    scaling_metric: Tuple[Literal["concurrency", "rps"], str],
    scale_down_delay: str = "0s",
    envs: Optional[Dict[str, str]] = None,
    resource_requests: Optional[Dict[str, str]] = None,
    resource_limits: Optional[Dict[str, str]] = None,
) -> Dict[str, Any]:
    """
    Creates a Knative Service with the specified configuration.

    This function creates a Knative Service with the provided service name, namespace, image,
    minimum and maximum instances, scaling metric, and scale down delay.

    The function validates the input parameters and raises a ValueError if any of them are invalid.

    Args:
        service_name (str): The name of the service.
        namespace (str): The namespace to create the service in.
        image (str): The Docker image for the service.
        min_instances (int): The minimum number of instances for the service.
        max_instances (int): The maximum number of instances for the service.
        scaling_metric (Tuple[Literal["concurrency", "rps"], str]): The scaling metric and target value.
        scale_down_delay (str, optional): The delay before scaling down. Defaults to "0s".
        resource_requests (Optional[Dict[str, str]], optional): The resource requests for the service, in the format {"cpu": "100m", "memory": "128Mi"}. Defaults to None.
        resource_limits (Optional[Dict[str, str]], optional): The resource limits for the service, in the format {"cpu": "200m", "memory": "256Mi", "gpu": "1"}. Defaults to None.

    Raises:
        ValueError: If any of the input parameters are invalid.

    Returns:
        None
    """
    if not isinstance(min_instances, int) or not isinstance(max_instances, int):
        raise ValueError("min_replicas and max_replicas must be integers")
    if min_instances > max_instances and not max_instances == 0:
        raise ValueError("min_replicas cannot be greater than max_replicas")

    metric_key, metric_value = scaling_metric
    if metric_key not in ["concurrency", "rps"]:
        raise ValueError(f"Invalid key in scaling_metric: {metric_key}")
    if not metric_value.isdigit():
        raise ValueError(f"Invalid value in scaling_metric: {metric_value}")

    config_map = client.V1ConfigMap(
        kind="ConfigMap",
        metadata=client.V1ObjectMeta(
            name="config-features", namespace="knative-serving"
        ),
        data={
            "kubernetes.podspec-tolerations": "enabled",
            "kubernetes.podspec-affinity": "enabled",
        },
    )

    apply_resource(config_map)

    container: Dict[str, Any] = {
        "image": image,
        # "imagePullPolicy": "Always",
        "command": shlex.split(entrypoint),
    }
    if envs:
        container["env"] = [
            {"name": key, "value": value} for key, value in envs.items()
        ]

    # Set a default value for resources if not provided
    container["resources"] = {
        "requests": {"cpu": "100m", "memory": "128Mi"},
        "limits": {"cpu": "200m", "memory": "256Mi"},
    }

    if resource_limits or resource_requests:
        container["resources"] = {}

    if resource_requests:
        if not all(key in VALID_RESOURCES for key in resource_requests.keys()):
            raise ValueError(
                f"Invalid resource request key. Valid keys are {VALID_RESOURCES}"
            )
        for resource, value in resource_requests.items():
            validate_resource(resource, value)
        container["resources"]["requests"] = resource_requests

    if resource_limits:
        if not all(
            key in (VALID_RESOURCES + VALID_RESOURCES_GPU)
            for key in resource_limits.keys()
        ):
            raise ValueError(
                f"Invalid resource limit key. Valid keys are {VALID_RESOURCES + VALID_RESOURCES_GPU}"
            )
        for resource, value in resource_limits.items():
            validate_resource(resource, value)
        container["resources"]["limits"] = resource_limits

    metric = {
        "autoscaling.knative.dev/metric": metric_key,
        "autoscaling.knative.dev/target": metric_value,
    }

    # Create the Knative Service
    knative_service = {
        "apiVersion": "serving.knative.dev/v1",
        "kind": "Service",
        "metadata": {
            "name": service_name,
            "namespace": namespace,
        },
        "spec": {
            "template": {
                "metadata": {
                    "annotations": {
                        "autoscaling.knative.dev/class": "kpa.autoscaling.knative.dev",
                        "autoscaling.knative.dev/min-scale": str(min_instances),
                        "autoscaling.knative.dev/max-scale": str(max_instances),
                        "autoscaling.knative.dev/scale-down-delay": scale_down_delay,
                        **metric,
                    },
                },
                "spec": {
                    "containers": [container],
                    "tolerations": [
                        {
                            "key": "app",
                            "operator": "Equal",
                            "value": "function",
                            "effect": "NoSchedule",
                        }
                    ],
                    "affinity": {
                        "nodeAffinity": {
                            "preferredDuringSchedulingIgnoredDuringExecution": [
                                {
                                    "weight": 100,
                                    "preference": {
                                        "matchExpressions": [
                                            {
                                                "key": "lifecycle",
                                                "operator": "In",
                                                "values": ["spot"],
                                            }
                                        ]
                                    },
                                },
                                {
                                    "weight": 50,
                                    "preference": {
                                        "matchExpressions": [
                                            {
                                                "key": "lifecycle",
                                                "operator": "In",
                                                "values": ["on-demand"],
                                            }
                                        ]
                                    },
                                },
                            ]
                        },
                    },
                },
            }
        },
    }

    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    try:
        service_resource.get(name=service_name, namespace=namespace)
        service_resource.patch(
            body=knative_service,
            namespace=namespace,
            name=service_name,
            content_type="application/merge-patch+json",
        )
    except ApiException as e:
        if e.status == 404:
            service_resource.create(body=knative_service, namespace=namespace)
        else:
            raise
    return knative_service


def list_knative_services(namespace: str) -> Any:
    """
    List all Knative Services in the specified namespace.

    Args:
        namespace (str): The namespace to list services in. Defaults to "knative-serving".

    Returns:
        None
    """
    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    return service_resource.get(namespace=namespace)


def delete_knative_service(service_name: str, namespace: str) -> None:
    """
    Delete a Knative Service.

    Args:
        service_name (str): The name of the service to delete.
        namespace (str): The namespace to delete the service from.

    Returns:
        None
    """
    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    service_resource.delete(name=service_name, namespace=namespace)


def list_knative_revisions(namespace: str, service_name: Optional[str] = None) -> Any:
    """
    List all Knative Revisions in the specified namespace.

    Args:
        namespace (str): The namespace to list revisions in.
        service_name (str, optional): The name of the service to list revisions for. If None, list revisions for all services.

    Returns:
        None
    """
    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    if not service_name:
        services = service_resource.get(namespace=namespace).items
    else:
        try:
            services = [service_resource.get(name=service_name, namespace=namespace)]
        except client.ApiException as e:
            if e.status == 404:
                services = []

    revision_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Revision"
    )

    try:
        revisions = revision_resource.get(namespace=namespace).items
    except client.ApiException as e:
        if e.status == 404:
            revisions = []

    for service in services:
        service_revisions = [
            rev
            for rev in revisions
            if rev.metadata.labels["serving.knative.dev/service"]
            == service.metadata.name
        ]

        # Add traffic information to the revisions
        traffic_list = (
            service.spec.traffic if service.spec and service.spec.traffic else []
        )
        for traffic in traffic_list:
            for rev in service_revisions:
                if rev.metadata.name == traffic.revisionName:
                    rev.traffic = f"{traffic.percent}%"

        return sorted(
            service_revisions,
            key=lambda x: int(
                x.metadata.labels["serving.knative.dev/configurationGeneration"]
            ),
            reverse=True,
        )


def split_traffic_among_revisions(
    namespace: str,
    service_name: str,
    traffic_splits: List[Tuple[str, int]],
    latest_revision_traffic: int,
) -> None:
    """
    Split traffic among the specified revisions of a service.

    Args:
        service_name (str): The name of the service.
        namespace (str): The namespace of the service.
        traffic_splits (List[Tuple[str, int]]): A list of tuples, where each tuple contains a revision name and a traffic percent.
        latest_revision_traffic (int): The traffic percent to assign to the latest revision.

    Raises:
        ValueError: If the traffic percent is not valid.

    Returns:
        None
    """
    total_traffic_percent = (
        sum(percent for _, percent in traffic_splits) + latest_revision_traffic
    )

    if not all(0 <= percent <= 100 for _, percent in traffic_splits) or not (
        0 <= latest_revision_traffic <= 100
    ):
        raise ValueError("All traffic percents must be between 0 and 100")

    if total_traffic_percent != 100:
        raise ValueError("Total traffic percent should be 100%")

    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    service = service_resource.get(name=service_name, namespace=namespace)

    traffic = [
        {
            "revisionName": revision_name,
            "percent": percent,
        }
        for revision_name, percent in traffic_splits
    ]

    if latest_revision_traffic > 0:
        traffic.append(
            {
                "latestRevision": True,
                "percent": latest_revision_traffic,
            }
        )

    service_spec = service.to_dict()

    service_spec["spec"]["traffic"] = traffic

    service_resource.patch(
        body=service_spec,
        namespace=namespace,
        name=service_name,
        content_type="application/merge-patch+json",
    )



import os
import platform
import shutil
from pathlib import Path

import requests

from paka.logger import logger
from paka.utils import download_url, get_project_data_dir

KUBECTL_VERSION_URL = "https://cdn.dl.k8s.io/release/stable.txt"
CHUNK_SIZE = 8192


def get_latest_kubectl_version() -> str:
    """Return the latest version of kubectl available for download."""
    try:
        response = requests.get(KUBECTL_VERSION_URL)
        response.raise_for_status()
        return response.text
    except requests.RequestException as e:
        logger.error(f"Failed to get latest kubectl version: {e}")
        return "v1.30.0"


# We are not pinning the version of kubectl to a specific version
# Get the latest version of kubectl should be safe
KUBECTL_VERSION = os.getenv("KUBECTL_VERSION", get_latest_kubectl_version())


# install_path is a full path to the kubectl binary
# It should be in a format like this /path/to/kubectl-x.xx.x/kubectl
def ensure_kubectl_by_path(install_path: Path) -> None:
    """Ensure kubectl is installed at the given path."""
    parent_dir = install_path.parent
    os.environ["PATH"] = f"{parent_dir.absolute()}{os.pathsep}{os.environ['PATH']}"

    if install_path.exists():
        return

    system = platform.system().lower()
    arch = platform.machine().lower()

    if arch in ["amd64", "x86_64"]:
        arch = "amd64"

    if arch not in ["amd64", "arm64"]:
        raise Exception(f"Unsupported architecture: {arch}")

    grandparent_dir = parent_dir.parent

    for old_kubectl_dir in grandparent_dir.glob("kubectl-*"):
        shutil.rmtree(old_kubectl_dir)

    if not install_path.exists():
        url = os.getenv(
            "KUBECTL_DOWNLOAD_URL",
            f"https://dl.k8s.io/release/{KUBECTL_VERSION}/bin/{system}/{arch}/kubectl",
        )
        if system == "windows" and not url.endswith(".exe"):
            url += ".exe"
        logger.info(f"Downloading {url}..")

        with download_url(url) as tmp_file:
            tmp_file_p = Path(tmp_file)
            tmp_file_p.chmod(0o755)
            parent_dir.mkdir(parents=True, exist_ok=True)
            shutil.copy2(tmp_file_p, install_path)


def ensure_kubectl() -> None:
    system = platform.system().lower()
    kubectl_path = (
        Path(get_project_data_dir()) / "bin" / f"kubectl-{KUBECTL_VERSION}" / "kubectl"
    )
    if system == "windows":
        kubectl_path = kubectl_path.with_suffix(".exe")

    ensure_kubectl_by_path(kubectl_path)



import pulumi_aws as aws

from paka.cluster.context import Context
from paka.utils import call_once


@call_once
def create_object_store(ctx: Context) -> None:
    """
    Creates an object store in AWS S3 based on the provided configuration.

    Returns:
        None
    """
    # `bucket` is the name of the bucket. It will avoid pulumi appending a random string to the name
    # `force_destroy`` is needed to delete the bucket when it's not empty
    bucket = aws.s3.Bucket(ctx.cluster_name, force_destroy=True)
    bucket.id.apply(lambda id: ctx.set_bucket(id))



import json

import pulumi
import pulumi_kubernetes as k8s
from kubernetes import client, config

from paka.cluster.context import Context


def create_namespace(ctx: Context, kubeconfig_json: str) -> None:
    # Pulumi does not support creating the default namespace again, so we need to handle it separately
    if ctx.namespace != "default":
        k8s.core.v1.Namespace(
            "app-ns",
            metadata={
                "name": ctx.namespace,
                "labels": {"istio-injection": "enabled"},
            },
            opts=pulumi.ResourceOptions(provider=ctx.k8s_provider),
        )
    else:
        config.load_kube_config_from_dict(json.loads(kubeconfig_json))
        # We are dealing with the default namespace
        api_instance = client.CoreV1Api()

        body = {"metadata": {"labels": {"istio-injection": "enabled"}}}

        api_instance.patch_namespace("default", body)



from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any

from pulumi import automation as auto

from paka.cluster.context import Context
from paka.cluster.pulumi import ensure_pulumi
from paka.config import CloudConfig, Config
from paka.constants import PULUMI_STACK_NAME
from paka.k8s.model_group.service import (
    cleanup_staled_model_group_services,
    create_model_group_service,
)
from paka.k8s.model_group.service_v1 import (
    create_model_group_service as create_model_group_service_v1,
)
from paka.logger import logger

STACK_NAME = "default"


class ClusterManager(ABC):
    """
    Abstract base class for a cluster manager.

    A ClusterManager is responsible for managing a cluster of compute resources.

    Subclasses must implement the abstract methods defined in this class.
    """

    config: Config
    cloud_config: CloudConfig

    def __init__(self, config: Config) -> None:
        self.config = config
        if not config.aws is None:
            self.cloud_config = config.aws
        self.ctx = Context()
        self.ctx.set_config(config)

    @abstractmethod
    def provision_k8s(self) -> None:
        pass

    def _stack_for_program(self, program: auto.PulumiFn) -> auto.Stack:
        return auto.create_or_select_stack(
            stack_name=PULUMI_STACK_NAME,
            project_name=self.cloud_config.cluster.name,
            program=program,
        )

    @cached_property
    def _stack(self) -> auto.Stack:
        ensure_pulumi()

        def program() -> None:
            self.provision_k8s()

        return self._stack_for_program(program)

    def create(self) -> None:
        if self.config.aws is None:
            raise ValueError("Only AWS is supported.")

        if self.config.aws:
            self._stack.set_config(
                "aws:region", auto.ConfigValue(value=self.cloud_config.cluster.region)
            )

        logger.info("Creating resources...")
        self._stack.up(on_output=logger.info)

        if (
            self.cloud_config.modelGroups is None
            and self.cloud_config.mixedModelGroups is None
        ):
            return

        namespace = self.cloud_config.cluster.namespace

        # Clean up staled model group resources before creating new ones
        model_group_names = [mg.name for mg in self.config.aws.modelGroups or []]
        mixed_model_group_names = [
            mg.name for mg in self.config.aws.mixedModelGroups or []
        ]
        all_group_names = model_group_names + mixed_model_group_names

        cleanup_staled_model_group_services(namespace, all_group_names)
        # TODO: We should clean up deployment as well

        for model_group in self.cloud_config.modelGroups or []:
            create_model_group_service(self.ctx, namespace, model_group)

        for mixed_model_group in self.cloud_config.mixedModelGroups or []:
            create_model_group_service_v1(self.ctx, namespace, mixed_model_group)

    def destroy(self) -> Any:
        logger.info("Destroying resources...")
        return self._stack.destroy(on_output=logger.info)

    def refresh(self) -> None:
        logger.info("Refreshing the stack...")
        self._stack.refresh(on_output=logger.info)

    def preview(self, *args: Any, **kwargs: Any) -> None:
        if not "on_output" in kwargs:
            kwargs["on_output"] = logger.info
        self._stack.preview(*args, **kwargs)



from __future__ import annotations

import os
from typing import Optional, Set

import boto3
import typer
from kubernetes import client
from tabulate import tabulate

from paka.cli.utils import (
    ensure_cluster_name,
    get_cluster_namespace,
    load_kubeconfig,
    read_pulumi_stack,
)
from paka.k8s.model_group.service import MODEL_PATH_PREFIX, filter_services
from paka.logger import logger

model_group_app = typer.Typer()


@model_group_app.command()
def list_downloaded_models(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
) -> None:
    """
    List all models that have been downloaded to the object store.
    """
    load_kubeconfig(cluster_name)
    cluster_name = ensure_cluster_name(cluster_name)
    bucket = read_pulumi_stack(cluster_name, "bucket")

    s3 = boto3.client("s3")
    response = s3.list_objects_v2(Bucket=bucket, Prefix=MODEL_PATH_PREFIX)
    if "Contents" in response:
        unique_models: Set[str] = set()
        for obj in response["Contents"]:
            key = obj["Key"]
            if key.startswith(f"{MODEL_PATH_PREFIX}/"):
                key = key[len(f"{MODEL_PATH_PREFIX}/") :]

            key = key.split("/")[0]
            unique_models.add(key)

        for key in unique_models:
            logger.info(key)
    else:
        logger.info("No models found.")


@model_group_app.command()
def list(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    ),
) -> None:
    """
    List all model groups.
    """
    load_kubeconfig(cluster_name)
    services = filter_services(get_cluster_namespace(cluster_name))

    # Get public model groups
    public_model_groups = [
        service.spec.selector.get("model", "")
        for service in services
        if service.spec
        and service.spec.selector
        and "model" in service.spec.selector
        and service.metadata
        and service.metadata.labels
        and (service.metadata.labels.get("is-public", "false") == "true")
    ]

    # Get private model groups
    private_model_groups = [
        service.spec.selector.get("model", "")
        for service in services
        if service.spec
        and service.spec.selector
        and "model" in service.spec.selector
        and service.metadata
        and service.metadata.labels
        and (service.metadata.labels.get("is-public", "false") == "false")
    ]

    model_groups = public_model_groups + private_model_groups

    v1 = client.CoreV1Api()
    cfg = v1.read_namespaced_config_map("config-domain", "knative-serving")
    cfg_data = cfg.data or {}
    filtered_keys = [key for key in cfg_data if key.endswith("sslip.io")]
    if not filtered_keys:
        if not model_groups:
            logger.info("No model groups found.")
        else:
            logger.info("\n".join(model_groups))
        return
    domain = filtered_keys[0]

    table = [(group, f"http://{group}.{domain}") for group in public_model_groups]
    table.extend([(group, f"private") for group in private_model_groups])
    logger.info(tabulate(table, headers=["Model Group", "Endpoint"]))



from __future__ import annotations

import threading
import time
from typing import List

import typer

from paka.cli.utils import load_cluster_manager, load_kubeconfig
from paka.k8s.utils import remove_crd_finalizers

cluster_app = typer.Typer()


@cluster_app.command()
def up(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
    no_kubeconfig: bool = typer.Option(
        False,
        "--no-kubeconfig",
        "-n",
        help="By default, the connection details of the newly created Kubernetes "
        "cluster are added to the default kubeconfig file (~/.kube/config). "
        "This allows kubectl to communicate with the new cluster. "
        "Use this option to prevent updating the kubeconfig file.",
    ),
) -> None:
    """
    Creates or updates a Kubernetes cluster based on the provided configuration.
    """
    cluster_manager = load_cluster_manager(cluster_config)
    cluster_manager.ctx.set_should_save_kubeconfig(not no_kubeconfig)
    cluster_manager.create()


@cluster_app.command()
def down(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Automatic yes to prompts. Use this option to bypass the confirmation "
        "prompt and directly proceed with the operation.",
    ),
) -> None:
    """
    Tears down the Kubernetes cluster, removing all associated resources and data.
    """
    if yes or typer.confirm(
        f"Are you sure you want to proceed with the operation? Please note that "
        "all resources and data will be permanently deleted.",
        default=False,
    ):
        cluster_manager = load_cluster_manager(cluster_config)

        # Sometime finalizers might block CRD deletion, so we need to force delete those.
        # This is best effort and might not work in all cases.
        # TODO: better way to handle this
        # https://github.com/kubernetes/kubernetes/issues/60538
        stop_event = threading.Event()

        def remove_finalizers_forever() -> None:
            try:
                crds = [
                    "scaledobjects.keda.sh",
                    "routes.serving.knative.dev",
                    "ingresses.networking.internal.knative.dev",
                ]

                load_kubeconfig(cluster_manager.cloud_config.cluster.name)

                while not stop_event.is_set():
                    for crd in crds:
                        try:
                            remove_crd_finalizers(crd)
                        except Exception as e:
                            pass
                    time.sleep(1)  # Wait for a second before the next iteration
            except:
                pass

        thread = threading.Thread(target=remove_finalizers_forever)
        thread.start()

        try:
            cluster_manager.destroy()
        finally:
            stop_event.set()
            thread.join()


@cluster_app.command()
def preview(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
    policy_packs: List[str] = typer.Option(
        [],
        "--policy-pack",
        "-p",
        help="Path to the policy pack.",
    ),
) -> None:
    """
    Previews the changes that will be applied to the cloud resources.
    """
    cluster_manager = load_cluster_manager(cluster_config)
    if policy_packs:
        cluster_manager.preview(policy_packs=policy_packs)
    else:
        cluster_manager.preview()


@cluster_app.command()
def refresh(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
) -> None:
    """
    Synchronize the local cluster state with the state in the cloud.
    """
    cluster_manager = load_cluster_manager(cluster_config)
    cluster_manager.refresh()



from __future__ import annotations

import json
import os
from typing import Optional

import typer

from paka.cli.utils import ensure_cluster_name
from paka.k8s.utils import update_kubeconfig
from paka.logger import logger
from paka.utils import read_pulumi_stack

kube_app = typer.Typer()


@kube_app.command()
def update(
    cluster_name: Optional[str] = typer.Option(
        os.getenv("PAKA_CURRENT_CLUSTER"),
        "--cluster",
        "-c",
        help="The name of the cluster.",
    )
) -> None:
    """
    Updates the default kubeconfig file (~/.kube/config) to include the connection
    details of the specified cluster.
    """
    logger.info("Updating kubeconfig...")
    cluster_name = ensure_cluster_name(cluster_name)
    kubeconfig = read_pulumi_stack(cluster_name, "kubeconfig")

    update_kubeconfig(json.loads(kubeconfig))
    logger.info("Successfully updated kubeconfig.")
